# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

import uuid
import time
import functools

from asyncio import iscoroutinefunction
from contextvars import ContextVar
from loguru import logger


TRACE_ID_CONTEXT: ContextVar[typing.Optional[str]] = ContextVar(r'trace_id', default=None)
TRACE_NO_CONTEXT: ContextVar[typing.Optional[int]] = ContextVar(r'trace_no', default=None)


def get_trace_id() -> str:

    return TRACE_ID_CONTEXT.get()


def refresh_trace_id(trace_id: typing.Optional[str] = None, trace_no: typing.Optional[str] = None) -> str:

    if trace_id is None:
        trace_id = str(uuid.uuid1())

    TRACE_ID_CONTEXT.set(trace_id)
    TRACE_NO_CONTEXT.set(trace_no)

    return trace_id


def inc_trace_no() -> int:

    trace_no = TRACE_NO_CONTEXT.get()
    trace_no = 0 if trace_no is None else trace_no + 1

    TRACE_NO_CONTEXT.set(trace_no)

    return trace_no


def trace_wrapper(func: typing.Callable) -> typing.Callable:

    if iscoroutinefunction(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):
            refresh_trace_id()
            return await func(*args, **kwargs)
    else:
        @functools.wraps(func)
        def _wrapper(*args, **kwargs):
            refresh_trace_id()
            return func(*args, **kwargs)

    return _wrapper


def tracing(func: typing.Callable) -> typing.Callable:

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):

        _time = time.time() * 1000

        try:
            return func(*args, **kwargs)
        finally:
            trace_time = r'{:.3f}'.format(time.time() * 1000 - _time)
            logger.bind(
                trace_id=get_trace_id(), trace_no=inc_trace_no(),
                trace_time=trace_time
            ).info(
                f'tracing <{func.__qualname__}> {trace_time}ms\n'
                f'args: {args}, kwargs: {kwargs}'
            )

    return _wrapper


def async_tracing(func: typing.Callable) -> typing.Callable:

    @functools.wraps(func)
    async def _wrapper(*args, **kwargs):

        _time = time.time() * 1000

        try:
            return await func(*args, **kwargs)
        finally:
            trace_time = r'{:.3f}'.format(time.time() * 1000 - _time)
            logger.bind(
                trace_id=get_trace_id(), trace_no=inc_trace_no(),
                trace_time=trace_time
            ).info(
                f'tracing <{func.__qualname__}> {trace_time}ms\n'
                f'args: {args}, kwargs: {kwargs}'
            )

    return _wrapper




# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import time
import typing
import signal
import socket
import asyncio
import base64
import msgpack

from ...extend.asyncio.base import Utils, install_uvloop
from ...extend.interface import RunnableInterface

from ..error import RouterError


DEFAULT_LIMIT = 0xffffff
DEFAULT_UNIX_SOCKET_ENDPOINT = r'/tmp/unix_socket_endpoint'


async def recv_msg(reader: 'asyncio.streams.StreamReader', timeout: typing.Optional[float] = None) -> typing.Any:

    result = None

    with Utils.suppress(asyncio.TimeoutError):

        data = await asyncio.wait_for(reader.readline(), timeout)

        if data:
            result = msgpack.loads(
                base64.b64decode(data)
            )

    return result


async def send_msg(writer: 'asyncio.streams.StreamWriter', data: typing.Any):

    writer.writelines(
        [
            base64.b64encode(
                msgpack.dumps(data)
            ),
            b'\n',
        ]
    )

    await writer.drain()


class UnixSocketServer:

    def __init__(self, client_connected_cb: typing.Callable, path: str, limit: int = DEFAULT_LIMIT):

        self._client_connected_cb: typing.Callable = client_connected_cb

        self._path: str = path
        self._limit: int = limit

        self._server: typing.Optional[asyncio.AbstractServer] = None

    async def open(self):

        if os.path.exists(self._path):
            os.remove(self._path)

        self._server = await asyncio.start_unix_server(
            self._client_connected_cb, self._path, limit=self._limit
        )

        await self._server.start_serving()

    async def close(self):

        if self._server is not None:
            self._server.close()
            self._server = None


class UnixSocketClient:

    def __init__(self, path: str, limit: int = DEFAULT_LIMIT):

        self._path: str = path
        self._limit: int = limit

        self._reader: typing.Optional['asyncio.streams.StreamReader'] = None
        self._writer: typing.Optional['asyncio.streams.StreamWriter'] = None

    @property
    def reader(self) -> 'asyncio.streams.StreamReader':
        return self._reader

    @property
    def writer(self) -> 'asyncio.streams.StreamWriter':
        return self._writer

    async def open(self):

        self._reader, self._writer = await asyncio.open_unix_connection(
            self._path,
            limit=self._limit
        )

    async def close(self):

        self._reader = None

        if self._writer is not None:
            self._writer.close()
            await self._writer.wait_closed()
            self._writer = None

    async def recv_msg(self, timeout: float = None) -> typing.Any:
        return await recv_msg(self._reader, timeout)

    async def send_msg(self, data: typing.Any):
        await send_msg(self._writer, data)


class Router:

    def __init__(self, root: str = r''):

        self._root: str = root
        self._reg_func: typing.Dict[str, typing.Callable] = {}

    def __repr__(self) -> str:

        return self._reg_func.__repr__()

    def __call__(self, method: str, *args, **kwargs) -> typing.Any:

        func = self._reg_func.get(method)

        if not func:
            raise RouterError(f'{method} not exists')

        return func(*args, **kwargs)

    def _reg(self, method: str, func: typing.Callable):

        _method = f'{self._root}{method}'

        if _method in self._reg_func:
            raise RouterError(f'{method} has exists')

        self._reg_func[_method] = func

    def reg(self, method) -> typing.Callable:

        def _reg_func(func):
            self._reg(method, func)
            return func

        return _reg_func

    def items(self) -> typing.Iterable[typing.Tuple[str, typing.Callable]]:

        return self._reg_func.items()

    def include(self, router: 'Router'):

        for method, func in router.items():
            self._reg(method, func)


class SocketConfig(typing.NamedTuple):
    client_connected_cb: typing.Callable
    address: typing.Any
    family: int
    backlog: typing.Optional[int]
    reuse_port: bool
    buffer_limit: int


class AsyncTcpServer(RunnableInterface):

    def __init__(
            self, client_connected_cb: typing.Callable, address: typing.Any, *,
            family: int = socket.AF_INET, backlog: typing.Optional[int] = None,
            reuse_port: bool = True, buffer_limit: int = DEFAULT_LIMIT,
            on_startup: typing.Callable = None, on_shutdown: typing.Callable = None
    ):

        self._listeners: typing.List[SocketConfig] = [
            SocketConfig(
                client_connected_cb, address,
                family, backlog, reuse_port, buffer_limit
            )
        ]

        self._on_startup: typing.Callable = on_startup
        self._on_shutdown: typing.Callable = on_shutdown

        self._servers: typing.List[asyncio.AbstractServer] = []

        signal.signal(signal.SIGINT, self._exit)
        signal.signal(signal.SIGTERM, self._exit)

    async def __aenter__(self):

        for server in self._servers:
            await server.__aenter__()

        return self

    async def __aexit__(self, exc_type, exc_value, _traceback):

        for server in self._servers:
            await server.__aexit__(exc_type, exc_value, _traceback)

    def add_listener(
            self, client_connected_cb: typing.Callable, address: typing.Any, *,
            family: int = socket.AF_INET, backlog: typing.Optional[int] = None,
            reuse_port: bool = True, buffer_limit: int = DEFAULT_LIMIT,
    ):

        self._listeners.append(
            SocketConfig(
                client_connected_cb, address,
                family, backlog, reuse_port, buffer_limit
            )
        )

    def run(self, *, debug=None):

        Utils.print_slogan()

        install_uvloop()

        asyncio.run(self._run(), debug=debug)

    async def _run(self):

        if self._on_startup is not None:
            await self._on_startup()

        for config in self._listeners:

            if config.family == socket.AF_UNIX:

                _socket_server = await asyncio.start_unix_server(
                    config.client_connected_cb, config.address, limit=config.buffer_limit
                )

                Utils.log.success(f'unix server [pid:{Utils.getpid()}] startup complete: {config.address}')

            else:

                _socket = socket.create_server(
                    config.address, family=config.family, backlog=config.backlog, reuse_port=config.reuse_port
                )

                _socket_server = await asyncio.start_server(
                    config.client_connected_cb,
                    limit=config.buffer_limit, sock=_socket
                )

                Utils.log.success(f'socket server [pid:{Utils.getpid()}] startup complete: {config.address}')

            self._servers.append(_socket_server)

        async with self:
            await asyncio.gather(*(_server.wait_closed() for _server in self._servers))

        if self._on_shutdown is not None:
            await self._on_shutdown()

    def _exit(self, *_):

        for _server in self._servers:
            _server.close()


class RobustConnection:

    def __init__(
            self, host: typing.Optional[str] = None, port: typing.Optional[str] = None, *,
            limit: int = DEFAULT_LIMIT, recycle: int = 86400, **kwargs
    ):

        self._host: typing.Optional[str] = host
        self._port: typing.Optional[int] = port
        self._limit: int = limit
        self._recycle: int = recycle
        self._setting: typing.Dict[str, typing.Any] = kwargs

        self._recycle_time: int = 0

        self._reader: typing.Optional[asyncio.StreamReader] = None
        self._writer: typing.Optional[asyncio.StreamWriter] = None

    @property
    def reader(self) -> asyncio.StreamReader:
        return self._reader

    @property
    def writer(self) -> asyncio.StreamWriter:
        return self._writer

    async def _check_recycle(self):

        if self._writer is not None and self._recycle_time > time.time():
            self._refresh_recycle_time()
        else:
            raise Exception(f'connection exceeds maximum idle time: {self._host}:{self._port}')

    def _refresh_recycle_time(self):
        self._recycle_time = int(time.time()) + self._recycle

    async def open(self):

        try:

            self._reader, self._writer = await asyncio.open_connection(
                self._host, self._port, limit=self._limit, **self._setting
            )

            self._refresh_recycle_time()

        except Exception as err:

            Utils.log.error(str(err))

    async def close(self):

        if self._writer is not None:

            try:
                self._writer.close()
                await self._writer.wait_closed()
            except Exception as err:
                Utils.log.error(str(err))

            self._recycle_time = 0

            self._reader = None
            self._writer = None

    async def reset(self):

        await self.close()
        await self.open()

    async def readline(self) -> bytes:

        try:
            await self._check_recycle()
            return await self._reader.readline()
        except Exception as err:
            await self.reset()
            raise err


    # noinspection SpellCheckingInspection
    async def readuntil(self, separator: bytes = b'\n') -> bytes:

        try:
            await self._check_recycle()
            return await self._reader.readuntil(separator)
        except Exception as err:
            await self.reset()
            raise err

    async def read(self, n: int = -1):

        try:
            await self._check_recycle()
            return await self._reader.read(n)
        except Exception as err:
            await self.reset()
            raise err

    async def write(self, data: bytes):

        try:
            await self._check_recycle()
            self._writer.write(data)
            await self._writer.drain()
        except Exception as err:
            await self.reset()
            raise err

    async def writelines(self, data: typing.Iterable[bytes]):

        try:
            await self._check_recycle()
            self._writer.writelines(data)
            await self._writer.drain()
        except Exception as err:
            await self.reset()
            raise err



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

from inspect import Signature, Parameter

# noinspection PyProtectedMember
from pydantic._internal._model_construction import ModelMetaclass as _ModelMetaclass
from pydantic.main import BaseModel as _BaseModel

from fastapi.params import Depends as _Depends


class ModelMetaclass(_ModelMetaclass):

    def __init__(
        cls,
        cls_name: str,
        bases: typing.Tuple[typing.Type[typing.Any], ...],
        namespace: typing.Dict[str, typing.Any]
    ):

        super().__init__(cls_name, bases, namespace)

        if cls.model_fields:

            # 原始参数类型清单
            annotations = typing.get_type_hints(cls)

            func_params = []

            for name, field in cls.model_fields.items():

                func_params.append(
                    Parameter(
                        name, Parameter.POSITIONAL_OR_KEYWORD,
                        default=field,
                        annotation=annotations.get(name, Parameter.empty)
                    )
                )

            # 更新类方法
            cls.params = lambda **kwargs: kwargs
            cls.params.__signature__ = Signature(func_params)


class BaseModel(_BaseModel, metaclass=ModelMetaclass):
    pass


class Depends(_Depends):

    def __init__(self, dependency: typing.Union[typing.Callable, BaseModel] = None, *, use_cache: bool = True):

        if isinstance(dependency, ModelMetaclass):
            super().__init__(dependency=dependency.params, use_cache=use_cache)
        else:
            super().__init__(dependency=dependency, use_cache=use_cache)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

from .base import Utils
from .error import catch_error


class M3U8:

    _DEFAULT_CONTENT = """#EXTM3U
#EXT-X-TARGETDURATION:{max_time}
#EXT-X-DISCONTINUITY
{content}
#EXT-X-ENDLIST"""

    def __init__(self, value: str, host: str= r''):

        self._host: str = host

        self._data: typing.List = self._parse(value)

    def __add__(self, other: 'M3U8') -> 'M3U8':

        shadow: 'M3U8' = Utils.deepcopy(self)
        shadow._data.extend(other._data)

        return shadow

    def __iadd__(self, other: 'M3U8') -> 'M3U8':

        self._data.extend(other._data)

        return self

    @staticmethod
    def _parse(value: str) -> typing.List:

        result = None

        with catch_error():

            record_list = []
            extinf_list = []

            for _line in value.strip().splitlines():

                if _line.startswith(r'#EXTINF:'):
                    extinf_list.append(float(_line.split(r':')[1].split(r',')[0]))
                elif not _line.startswith(r'#'):
                    record_list.append(_line)

            if len(extinf_list) == len(record_list):
                result = zip(record_list, extinf_list)
            else:
                raise Exception(r'parse m3u8 content error')

        return result

    def split(self, _start: int, _end: int) -> str:

        timer = 0
        max_time = 0

        infos = []

        for _url, _time in self._data:

            max_time = max(max_time, Utils.math.ceil(_time))

            _next = timer + _time

            if timer < _start and _next < _start:
                timer = _next
                continue
            elif timer > _end:
                break

            infos.append(f'#EXTINF:{_time}\n{self._host}{_url}')
            timer = _next

        return self._DEFAULT_CONTENT.format(max_time=max_time, content='\n'.join(infos))



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import sys
import time
import signal
import typing

from multiprocessing import Process, set_start_method
from multiprocessing.shared_memory import SharedMemory

from .base import Utils
from .interface import ContextManager, RunnableInterface
from .struct import ByteArrayAbstract


def set_spawn_mode():

    set_start_method(r'spawn')


def fork_processes() -> int:

    pid = os.fork()

    if pid == 0:
        Utils.urandom_seed()

    return pid


class Daemon(RunnableInterface):

    def __init__(
            self, target: typing.Callable, sub_process_num: int, *,
            set_affinity: typing.Optional[typing.List[int]] = None, join_timeout: int = 10, **kwargs
    ):

        self._target: typing.Callable = target
        self._kwargs: typing.Dict = kwargs

        self._sub_process: typing.Set = set()
        self._sub_process_num: int = sub_process_num

        self._set_affinity: typing.Optional[typing.List[int]] = set_affinity
        self._join_timeout: int = join_timeout

        signal.signal(signal.SIGINT, self._kill_process)
        signal.signal(signal.SIGTERM, self._kill_process)

        if self._set_affinity:

            if sub_process_num > len(self._set_affinity):
                raise Exception(f'process num {sub_process_num}, affinity {self._set_affinity}')

            cpu_count = os.cpu_count()

            for _core in self._set_affinity:
                if _core >= cpu_count:
                    raise Exception(f'cpu cores num {cpu_count}, cores {self._set_affinity}')

    def _kill_process(self, *_):

        for process in self._sub_process:
            os.kill(process.ident, signal.SIGINT)

        for process in self._sub_process:
            process.join(self._join_timeout)
            process.kill()

    def _fill_process(self):

        for idx in range(self._sub_process_num - len(self._sub_process)):

            process = Process(target=self._target, args=(idx,), kwargs=self._kwargs)
            process.start()

            self._sub_process.add(process)

        if self._set_affinity:

            for _idx, _process in enumerate(self._sub_process):

                _pid = _process.pid
                _cores = self._set_affinity[_idx: _idx + 1]

                os.sched_setaffinity(_pid, _cores)
                sys.stdout.write(f'process {_pid} affinity {os.sched_getaffinity(_pid)}\n')

    def _check_process(self):

        for process in self._sub_process.copy():

            if process.is_alive():
                continue

            self._sub_process.remove(process)
            sys.stderr.write(f'kill process {process.ident}\n')

        return len(self._sub_process) > 0

    def run(self):

        self._fill_process()

        while self._check_process():
            time.sleep(1)


class SharedByteArray(ByteArrayAbstract, ContextManager):

    def __init__(self, name: typing.Optional[str] = None, create: bool = False, size: int = 0):

        ByteArrayAbstract.__init__(self)

        self._shared_memory: SharedMemory = SharedMemory(name, create, size)
        self._create_mode: bool = create

    def _context_release(self):

        self.release()

    def release(self):

        self._shared_memory.close()

        if self._create_mode:
            self._shared_memory.unlink()

    def read(self, size: int) -> memoryview:
        return self._shared_memory.buf[:size]

    def write(self, buffer: memoryview):
        self._shared_memory.buf[:len(buffer)] = buffer


class HeartbeatChecker(ContextManager):

    def __init__(self, name: str = r'default', timeout: int = 60):

        self._name: str = f'heartbeat_{name}'

        self._timeout: int = timeout

        self._byte_array: typing.Optional[SharedByteArray] = None

        # noinspection PyBroadException
        try:
            self._byte_array = SharedByteArray(self._name, True, 8)
        except Exception as _:
            self._byte_array = SharedByteArray(self._name)

    def _context_release(self):

        self.release()

    @property
    def refresh_time(self) -> int:

        if self._byte_array is not None:
            return self._byte_array.read_unsigned_int()
        else:
            return 0

    def release(self):

        if self._byte_array is not None:
            self._byte_array.release()
            self._byte_array = None

    def check(self) -> bool:

        return (Utils.timestamp() - self.refresh_time) < self._timeout

    def refresh(self):

        if self._byte_array is not None:
            self._byte_array.write_unsigned_int(Utils.timestamp())



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import time
import typing

import asyncio

from abc import abstractmethod

from ntplib import NTPClient

from .base import Utils
from .task import IntervalTask, MultiTasks

from ..error import NTPCalibrateError
from ..interface import TaskInterface


class _Interface(TaskInterface):
    """NTP客户端接口定义
    """

    @abstractmethod
    def start(self):
        raise NotImplementedError()

    @abstractmethod
    def stop(self):
        raise NotImplementedError()

    @abstractmethod
    def is_running(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def calibrate_offset(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def offset(self) -> float:
        raise NotImplementedError()

    @property
    @abstractmethod
    def timestamp(self) -> float:
        raise NotImplementedError()


class AsyncNTPClient(_Interface):
    """异步NTP客户端类
    """

    @classmethod
    async def create(cls, host: str) -> 'AsyncNTPClient':

        client = cls(host)

        await client.calibrate_offset()

        client.start()

        return client

    def __init__(
            self, host: str, *,
            version: int = 2, port: str = r'ntp',
            timeout: int = 5, interval: int = 3600, sampling: int = 5
    ):

        self._settings = {
            r'host': host,
            r'version': version,
            r'port': port,
            r'timeout': timeout,
        }

        self._client: NTPClient = NTPClient()
        self._offset: float = 0

        self._sync_task: IntervalTask = IntervalTask.create(interval, self.calibrate_offset)

        self._sampling: int = sampling

    def start(self):
        return self._sync_task.start()

    def stop(self):
        return self._sync_task.stop()

    def is_running(self) -> bool:
        return self._sync_task.is_running()

    async def calibrate_offset(self):
        return await asyncio.to_thread(self._calibrate_offset)

    def _calibrate_offset(self):

        samples = []
        host_name = self._settings[r'host']

        # 多次采样取中位数，减少抖动影响
        for _ in range(self._sampling):
            try:
                resp = self._client.request(**self._settings)
                samples.append(resp.offset)
            except Exception as err:
                Utils.log.warning(f'NTP server {host_name} request error: {err}')

        if samples:
            self._offset = float(sum(samples) / len(samples))
            Utils.log.debug(f'NTP server {host_name} offset median {self._offset} samples: {samples}')
        else:
            raise NTPCalibrateError(f'NTP server {host_name} not available, timestamp uncalibrated')

    @property
    def offset(self) -> float:
        return self._offset

    @property
    def timestamp(self) -> float:

        return time.time() + self._offset


class AsyncNTPClientPool(_Interface):
    """异步NTP客户端池，多节点实现高可用
    """

    @classmethod
    async def create(cls, hosts: typing.List[str]) -> 'AsyncNTPClientPool':

        client_pool = cls()

        for host in hosts:
            client_pool.append(host)

        await client_pool.calibrate_offset()

        client_pool.start()

        return client_pool

    def __init__(self):

        self._clients: typing.List[AsyncNTPClient] = []
        self._running: bool = False

    def append(self, host, *, version=2, port='ntp', timeout=5, interval=3600, sampling=5):

        client = AsyncNTPClient(host, version=version, port=port, timeout=timeout, interval=interval, sampling=sampling)

        self._clients.append(client)

        if self._running:
            client.start()

    def start(self):

        for client in self._clients:
            client.start()

        self._running = True

    def stop(self):

        for client in self._clients:
            client.stop()

        self._running = False

    def is_running(self) -> bool:
        return self._running

    async def calibrate_offset(self):

        tasks = MultiTasks()

        for client in self._clients:
            tasks.append(client.calibrate_offset())

        await tasks

    @property
    def offset(self) -> float:

        samples = []

        for client in self._clients:
            samples.append(client.offset)

        return float(sum(samples) / len(samples))

    @property
    def timestamp(self) -> float:

        return time.time() + self.offset



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

from abc import ABCMeta, abstractmethod

from .base import Utils
from .interface import ContextManager


class TransactionAbstract(metaclass=ABCMeta):
    """事务接口
    """

    def __init__(
            self, *,
            commit_callback: typing.Optional[typing.Callable] = None,
            rollback_callback: typing.Optional[typing.Callable] = None
    ):

        self._commit_callbacks: typing.List[typing.Callable] = []
        self._rollback_callbacks: typing.List[typing.Callable] = []

        if commit_callback is not None:
            self._commit_callbacks.append(commit_callback)

        if rollback_callback is not None:
            self._rollback_callbacks.append(rollback_callback)

    def _clear_callbacks(self):

        self._commit_callbacks.clear()
        self._rollback_callbacks.clear()

        self._commit_callbacks = self._rollback_callbacks = None

    def add_commit_callback(self, _callable: typing.Callable, *args, **kwargs):

        if self._commit_callbacks is None:
            return

        if args or kwargs:
            self._commit_callbacks.append(
                Utils.func_partial(_callable, *args, **kwargs)
            )
        else:
            self._commit_callbacks.append(_callable)

    def add_rollback_callback(self, _callable: typing.Callable, *args, **kwargs):

        if self._rollback_callbacks is None:
            return

        if args or kwargs:
            self._rollback_callbacks.append(
                Utils.func_partial(_callable, *args, **kwargs)
            )
        else:
            self._rollback_callbacks.append(_callable)

    @abstractmethod
    def commit(self):

        raise NotImplementedError()

    @abstractmethod
    def rollback(self):

        raise NotImplementedError()

    def bind(self, trx: 'TransactionAbstract'):

        self.add_commit_callback(trx.commit)
        self.add_rollback_callback(trx.rollback)


class Transaction(TransactionAbstract, ContextManager):
    """事务对象

    使用上下文实现的一个事务对象，可以设置commit和rollback回调
    未显示commit的情况下，会自动rollback

    """

    def _context_release(self):

        self.rollback()

    def commit(self):

        if self._commit_callbacks is None:
            return

        callbacks = self._commit_callbacks.copy()

        self._clear_callbacks()

        for _callable in callbacks:
            try:
                _callable()
            except Exception as err:
                Utils.log.critical(f'transaction commit error:\n{err}')

    def rollback(self):

        if self._rollback_callbacks is None:
            return

        callbacks = self._rollback_callbacks.copy()

        self._clear_callbacks()

        for _callable in callbacks:
            try:
                _callable()
            except Exception as err:
                Utils.log.critical(f'transaction rollback error:\n{err}')



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import aiosmtplib

from email.message import EmailMessage

from .base import Utils


class EmailBody:

    def __init__(self, sender: str, recipients: typing.Sequence[str], message: EmailMessage):

        self._sender: str = sender
        self._recipients: typing.Sequence[str] = recipients
        self._message: EmailMessage = message

    @property
    def sender(self) -> str:
        return self._sender

    @property
    def recipients(self) -> typing.Sequence[str]:
        return self._recipients

    @property
    def message(self) -> EmailMessage:
        return self._message

    @staticmethod
    def create_message(
            sender: str, recipients: typing.Sequence[str], subject: str,
            content: str, content_type: str = r'text/html'
    ) -> EmailMessage:

        message = EmailMessage()

        if not sender:
            message[r'From'] = sender

        if not recipients:
            message[r'To'] = recipients

        message[r'Subject'] = subject
        message.set_content(content)
        message.set_type(content_type)

        return message


class SMTPClient:

    def __init__(self, username: str, password: str, hostname: str, port: int, retry_count: int = 5, **kwargs):

        self._username: str = username
        self._password: str = password

        self._hostname: str = hostname
        self._port: int = port

        self._retry_count: int = retry_count

        self._smtp_settings: typing.Dict = kwargs

    @staticmethod
    def format_address(nickname: str, mailbox: str) -> str:

        return f'{nickname}<{mailbox}>'

    def format_addresses(self, addresses: typing.List[typing.Tuple[str, str]]) -> str:

        return r';'.join(self.format_address(*addr) for addr in addresses)

    # 发送多封邮件
    async def send_messages(self, email_body_list: typing.List[EmailBody]):

        resp = None

        try:

            async with aiosmtplib.SMTP(hostname=self._hostname, port=self._port, **self._smtp_settings) as client:

                await client.login(self._username, self._password)

                for email_body in email_body_list:

                    resp = await client.send_message(
                        email_body.message, sender=email_body.sender, recipients=email_body.recipients
                    )

                    Utils.log.info(f'{email_body.recipients} => {resp}')

        except aiosmtplib.SMTPException as err:

            Utils.log.error(err)

        return resp

    # 发送单封邮件
    async def send_message(self, sender: str, recipients: typing.Sequence[str], message: EmailMessage):

        return await self.send_messages(
            [EmailBody(sender, recipients, message)]
        )

    # 简单的发送
    async def send(self, sender: str, recipients: typing.Sequence[str], subject: str, content: str):

        return await self.send_message(
            sender, recipients,
            EmailBody.create_message(sender, recipients, subject, content)
        )



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import asyncio
import grpc
import msgpack

from abc import abstractmethod
from collections import OrderedDict

from ...extend.base import Utils


CHANNEL_USABLE_STATE = (grpc.ChannelConnectivity.READY, grpc.ChannelConnectivity.IDLE)


class GRPCClientAbstract:

    def __init__(
            self, *,
            credentials: typing.Optional[grpc.ChannelCredentials] = None,
            options: typing.Optional[grpc.aio.ChannelArgumentType] = None,
            compression: typing.Optional[grpc.Compression] = None,
            interceptors: typing.Optional[typing.Sequence[grpc.aio.ClientInterceptor]] = None,
            request_serializer: typing.Callable = msgpack.dumps,
            response_deserializer: typing.Callable = msgpack.loads
    ):

        self._credentials: typing.Optional[grpc.ChannelCredentials] = credentials
        self._options: typing.Optional[grpc.aio.ChannelArgumentType] = options
        self._compression: typing.Optional[grpc.Compression] = compression
        self._interceptors: typing.Optional[typing.Sequence[grpc.aio.ClientInterceptor]] = interceptors

        self._request_serializer: typing.Callable = request_serializer
        self._response_deserializer: typing.Callable = response_deserializer

    async def _make_channel(self, target: str, timeout: int = 0) -> grpc.aio.Channel:

        channel = None

        try:

            if self._credentials is None:
                channel = grpc.aio.insecure_channel(
                    target, self._options, self._compression, self._interceptors
                )
            else:
                channel = grpc.aio.secure_channel(
                    target, self._credentials, self._options, self._compression, self._interceptors
                )

            if timeout > 0:
                await asyncio.wait_for(channel.channel_ready(), timeout)

        except asyncio.TimeoutError as err:

            await channel.close()

            raise err

        return channel

    @abstractmethod
    async def open(self, targets: typing.List[str], timeout: int = 0):

        raise NotImplementedError()

    @abstractmethod
    async def close(self):

        raise NotImplementedError()

    @abstractmethod
    async def get_channel(self, *args, **kwargs) -> grpc.aio.Channel:

        raise NotImplementedError()

    @abstractmethod
    async def reset_channel(self, targets, timeout=0):

        raise NotImplementedError()

    async def unary_unary(
            self, channel: grpc.aio.Channel,
            method: str, call_params: typing.Dict
    ) -> grpc.aio.UnaryUnaryCall:

        return await channel.unary_unary(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(call_params)

    async def unary_stream(
            self, channel: grpc.aio.Channel,
            method: str, call_params: typing.Dict
    ) -> grpc.aio.UnaryStreamCall:

        return channel.unary_stream(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(call_params)

    async def stream_unary(
            self, channel: grpc.aio.Channel,
            method: str, call_params: typing.Union[typing.Iterable[typing.Dict], typing.AsyncIterable[typing.Dict]]
    ) -> grpc.aio.StreamUnaryCall:

        return await channel.stream_unary(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(call_params)

    async def stream_stream(
            self, channel: grpc.aio.Channel,
            method: str, call_params: typing.Union[typing.Iterable[typing.Dict], typing.AsyncIterable[typing.Dict]]
    ) -> grpc.aio.StreamStreamCall:

        return channel.stream_stream(
            method,
            request_serializer=self._request_serializer,
            response_deserializer=self._response_deserializer,
        )(call_params)


class GRPCClient(GRPCClientAbstract):

    def __init__(
            self, *,
            credentials: typing.Optional[grpc.ChannelCredentials] = None,
            options: typing.Optional[grpc.aio.ChannelArgumentType] = None,
            compression: typing.Optional[grpc.Compression] = None,
            interceptors: typing.Optional[typing.Sequence[grpc.aio.ClientInterceptor]] = None,
            request_serializer: typing.Callable = msgpack.dumps,
            response_deserializer: typing.Callable = msgpack.loads
    ):

        super().__init__(
            credentials=credentials, options=options, compression=compression, interceptors=interceptors,
            request_serializer=request_serializer, response_deserializer=response_deserializer
        )

        self._lock: asyncio.Lock = asyncio.Lock()
        self._channels: OrderedDict = OrderedDict()

    async def open(self, targets: typing.List[str], timeout: int = 0):

        await self.reset_channel(targets, timeout)

    async def close(self):

        async with self._lock:

            for _channel in self._channels.values():
                await _channel.close()

            self._channels.clear()

    async def get_channel(self) -> grpc.aio.Channel:

        channel = None

        async with self._lock:

            for _ in range(len(self._channels)):

                _target, _channel = self._channels.popitem()
                self._channels[_target] = _channel

                if _channel.get_state() in CHANNEL_USABLE_STATE:
                    channel = _channel
                    break
                else:
                    Utils.log.warning(f'grpc server unusable: {_target}')

            else:

                Utils.log.error(f'all grpc server unusable: {list(self._channels.keys())}')

        return channel

    async def reset_channel(self, targets: typing.List[str], timeout: int = 0):

        async with self._lock:

            if len(self._channels) > 0:

                for _channel in self._channels.values():
                    await _channel.close()

                self._channels.clear()

            for target in set(targets):

                channel = await self._make_channel(target, timeout)

                if channel is not None:
                    self._channels[target] = channel


class GRPCClientSlots(GRPCClientAbstract):

    def __init__(
            self, *,
            credentials: typing.Optional[grpc.ChannelCredentials] = None,
            options: typing.Optional[grpc.aio.ChannelArgumentType] = None,
            compression: typing.Optional[grpc.Compression] = None,
            interceptors: typing.Optional[typing.Sequence[grpc.aio.ClientInterceptor]] = None,
            request_serializer: typing.Callable = msgpack.dumps,
            response_deserializer: typing.Callable = msgpack.loads
    ):

        super().__init__(
            credentials=credentials, options=options, compression=compression, interceptors=interceptors,
            request_serializer=request_serializer, response_deserializer=response_deserializer
        )

        self._lock: asyncio.Lock = asyncio.Lock()
        self._channels: typing.List[grpc.Channel] = []

    async def open(self, targets, timeout=None):

        await self.reset_channel(targets, timeout)

    async def close(self):

        async with self._lock:

            for _channel in self._channels:
                await _channel.close()

            self._channels.clear()

    async def get_channel(self, name) -> grpc.aio.Channel:

        async with self._lock:
            index = Utils.md5_u64(str(name)) % len(self._channels)
            return self._channels[index]

    async def reset_channel(self, targets: typing.List[str], timeout: int = 0):

        async with self._lock:

            if len(self._channels) > 0:

                for _channel in self._channels:
                    await _channel.close()

                self._channels.clear()

            for target in set(targets):

                channel = await self._make_channel(target, timeout)

                if channel is not None:
                    self._channels.append(channel)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

from ..cache import StackCache
from ..error import catch_error

from .net import HTTPClientPool, http_types
from .future import ThreadPool


class FileLoader:
    """带缓存的网络文件加载器
    """

    def __init__(self, maxsize: int = 0xff, ttl: int = 3600, thread: int = 32):

        self._cache: StackCache = StackCache(maxsize, ttl)

        self._thread_pool: ThreadPool = ThreadPool(thread)
        self._http_client: HTTPClientPool = HTTPClientPool(limit=thread)

    @staticmethod
    def _read(file: str) -> typing.AnyStr:

        with open(file, r'rb') as stream:
            return stream.read()

    async def read(self, file: str) -> typing.AnyStr:

        with catch_error():

            if self._cache.has(file):
                return self._cache.get(file)

            result = await self._thread_pool.run(self._read, file)

            self._cache.set(file, result)

            return result

    async def fetch(
            self, url: str, params: typing.Optional[http_types.QueryParamTypes] = None, *,
            cookies: typing.Optional[http_types.CookieTypes] = None,
            headers: typing.Optional[http_types.HeaderTypes] = None
    ) -> typing.AnyStr:

        with catch_error():

            if self._cache.has(url):
                return self._cache.get(url)

            result = await self._http_client.get(url, params=params, cookies=cookies, headers=headers)

            self._cache.set(url, result)

            return result



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing

from cachetools import TTLCache

from .base import Utils
from .transaction import Transaction


class StackCache:
    """堆栈缓存

    使用运行内存作为高速缓存，可有效提高并发的处理能力

    """

    def __init__(self, maxsize: float = 0xff, ttl: float = 60):

        self._cache: TTLCache = TTLCache(maxsize, ttl)

    def has(self, key: str) -> bool:

        return key in self._cache

    def get(self, key: str, default: typing.Any = None) -> typing.Any:

        return self._cache.get(key, default)

    def set(self, key: str, val: typing.Any):

        self._cache[key] = val

    def incr(self, key: str, val: float = 1) -> float:

        res = self.get(key, 0) + val

        self.set(key, res)

        return res

    def decr(self, key: str, val: float = 1) -> float:

        res = self.get(key, 0) - val

        self.set(key, res)

        return res

    def delete(self, key: str):

        del self._cache[key]

    def size(self) -> int:

        return len(self._cache)

    def clear(self):

        return self._cache.clear()


class PeriodCounter:

    def __init__(self, time_slice: float = 60, key_prefix: str = r'', maxsize: float = 0xffff):

        self._time_slice: float = time_slice
        self._key_prefix: str = key_prefix

        # 缓存对象初始化，key最小过期时间60秒
        self._cache: StackCache = StackCache(maxsize, time_slice)

    def _get_key(self, key: typing.Optional[str] = None) -> str:

        time_period = Utils.math.floor(Utils.timestamp() / self._time_slice)

        if key is None:
            return f'{self._key_prefix}_{time_period}'
        else:
            return f'{self._key_prefix}_{key}_{time_period}'

    def get(self, key: typing.Optional[str] = None) -> typing.Any:

        _key = self._get_key(key)

        return self._cache.get(_key, 0)

    def incr(self, val: float, key: typing.Optional[str] = None) -> float:

        _key = self._get_key(key)

        return self._cache.incr(_key, val)

    def incr_with_trx(self, val: float, key: typing.Optional[str] = None) -> typing.Tuple[float, Transaction]:

        _key = self._get_key(key)

        trx = Transaction()
        trx.add_rollback_callback(self._cache.decr, _key, val)

        return self._cache.incr(_key, val), trx

    def decr(self, val: float, key: typing.Optional[str] = None) -> float:

        _key = self._get_key(key)

        return self._cache.decr(_key, val)

    def decr_with_trx(self, val: float, key: typing.Optional[str] = None) -> typing.Tuple[float, Transaction]:

        _key = self._get_key(key)

        trx = Transaction()
        trx.add_rollback_callback(self._cache.incr, _key, val)

        return self._cache.decr(_key, val), trx



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import httpx

from contextlib import contextmanager

# noinspection PyProtectedMember
from httpx import _types as http_types

from .base import Utils, AsyncCirculatoryForSecond


DEFAULT_RETRIES = 5
DEFAULT_TIMEOUT = httpx.Timeout(timeout=60)
DEFAULT_LIMITS = httpx.Limits(max_connections=200, max_keepalive_connections=50)
DOWNLOAD_TIMEOUT = httpx.Timeout(timeout=600)


@contextmanager
def _catch_error():
    """异常捕获

    通过with语句捕获异常，代码更清晰，还可以使用Ignore异常安全的跳出with代码块

    """

    try:

        yield

    except httpx.HTTPStatusError as err:

        Utils.log.warning(err)

    except Exception as err:

        Utils.log.error(err)


class Result(dict):

    def __init__(self, status: int, headers: typing.Dict, cookies: typing.Dict, body: typing.Any):

        super().__init__(status=status, headers=headers, cookies=cookies, body=body)

    def __bool__(self):

        return (self.status >= 200) and (self.status <= 299)

    @property
    def status(self) -> int:

        return self.get(r'status')

    @property
    def headers(self) -> typing.Dict:

        return self.get(r'headers')

    @property
    def cookies(self) -> typing.Dict:

        return self.get(r'cookies')

    @property
    def body(self) -> typing.Any:

        return self.get(r'body')


class _ClientBase:

    def __init__(
            self,
            verify: http_types.VerifyTypes = False,
            limits: httpx.Limits = DEFAULT_LIMITS,
            timeout: http_types.TimeoutTypes = DOWNLOAD_TIMEOUT,
            **kwargs
    ):

        self._client: httpx.AsyncClient = httpx.AsyncClient(verify=verify, limits=limits, timeout=timeout, **kwargs)

    async def close(self):

        if not self._client.is_closed:
            await self._client.aclose()


class HTTPClientPool(_ClientBase):
    """HTTP客户端类
    """

    def __init__(
            self,
            retries: int = DEFAULT_RETRIES,
            verify: http_types.VerifyTypes = False,
            limits: httpx.Limits = DEFAULT_LIMITS,
            timeout: http_types.TimeoutTypes = DEFAULT_TIMEOUT,
            **kwargs
    ):

        self._retries: int = retries

        super().__init__(verify=verify, limits=limits, timeout=timeout, **kwargs)

    def _handle_response(self, response: httpx.Response) -> bytes:

        return response.content

    async def send_request(
            self,
            method: str,
            url: http_types.URLTypes,
            *,
            content: typing.Optional[http_types.RequestContent] = None,
            data: typing.Optional[http_types.RequestData] = None,
            files: typing.Optional[http_types.RequestFiles] = None,
            json: typing.Optional[typing.Any] = None,
            params: typing.Optional[http_types.QueryParamTypes] = None,
            headers: typing.Optional[http_types.HeaderTypes] = None,
            cookies: typing.Optional[http_types.CookieTypes] = None,
            **kwargs
    ) -> Result:

        response = None

        async for times in AsyncCirculatoryForSecond(max_times=self._retries):

            try:

                _resp = await self._client.request(
                    method, url,
                    content=content, data=data, files=files, json=json, params=params,
                    headers=headers, cookies=cookies,
                    **kwargs
                )

                response = Result(
                    _resp.status_code,
                    dict(_resp.headers),
                    dict(_resp.cookies),
                    self._handle_response(_resp)
                )

            except httpx.HTTPStatusError as err:

                # 重新尝试的话，会记录异常，否则会继续抛出异常

                if err.response.status_code < 500:
                    raise err
                elif times >= self._retries:
                    raise err
                else:
                    Utils.log.warning(err)
                    continue

            except (httpx.TimeoutException, httpx.NetworkError) as err:

                if times >= self._retries:
                    raise err
                else:
                    Utils.log.warning(err)
                    continue

            else:

                Utils.log.info(f'{method} {url} => status:{response.status}')
                break

            finally:

                if times > 1:
                    Utils.log.warning(f'{method} {url} => retry:{times}')

        return response

    async def close(self):

        if not self._client.is_closed:
            await self._client.aclose()

    async def get(
            self,
            url: http_types.URLTypes,
            *,
            params: typing.Optional[http_types.QueryParamTypes] = None,
            headers: typing.Optional[http_types.HeaderTypes] = None,
            cookies: typing.Optional[http_types.CookieTypes] = None
    ) -> typing.Any:

        with _catch_error():

            resp = await self.send_request(r'GET', url, params=params, headers=headers, cookies=cookies)

            return resp.body

    async def options(
            self,
            url: http_types.URLTypes,
            *,
            params: typing.Optional[http_types.QueryParamTypes] = None,
            headers: typing.Optional[http_types.HeaderTypes] = None,
            cookies: typing.Optional[http_types.CookieTypes] = None
    ) -> typing.Any:

        with _catch_error():

            resp = await self.send_request(r'OPTIONS', url, params=params, headers=headers, cookies=cookies)

            return resp.headers

    async def head(
            self,
            url: http_types.URLTypes,
            *,
            params: typing.Optional[http_types.QueryParamTypes] = None,
            headers: typing.Optional[http_types.HeaderTypes] = None,
            cookies: typing.Optional[http_types.CookieTypes] = None
    ) -> typing.Any:

        with _catch_error():

            resp = await self.send_request(r'HEAD', url, params=params, headers=headers, cookies=cookies)

            return resp.headers

    async def post(
            self,
            url: http_types.URLTypes,
            *,
            content: typing.Optional[http_types.RequestContent] = None,
            data: typing.Optional[http_types.RequestData] = None,
            files: typing.Optional[http_types.RequestFiles] = None,
            json: typing.Optional[typing.Any] = None,
            params: typing.Optional[http_types.QueryParamTypes] = None,
            headers: typing.Optional[http_types.HeaderTypes] = None,
            cookies: typing.Optional[http_types.CookieTypes] = None
    ) -> typing.Any:

        with _catch_error():

            resp = await self.send_request(
                r'POST', url,
                content=content, data=data, files=files, json=json, params=params,
                headers=headers, cookies=cookies
            )

            return resp.body

    async def put(
            self,
            url: http_types.URLTypes,
            *,
            content: typing.Optional[http_types.RequestContent] = None,
            data: typing.Optional[http_types.RequestData] = None,
            files: typing.Optional[http_types.RequestFiles] = None,
            json: typing.Optional[typing.Any] = None,
            params: typing.Optional[http_types.QueryParamTypes] = None,
            headers: typing.Optional[http_types.HeaderTypes] = None,
            cookies: typing.Optional[http_types.CookieTypes] = None
    ) -> typing.Any:

        with _catch_error():

            resp = await self.send_request(
                r'PUT', url,
                content=content, data=data, files=files, json=json, params=params,
                headers=headers, cookies=cookies
            )

            return resp.body

    async def patch(
            self,
            url: http_types.URLTypes,
            *,
            content: typing.Optional[http_types.RequestContent] = None,
            data: typing.Optional[http_types.RequestData] = None,
            files: typing.Optional[http_types.RequestFiles] = None,
            json: typing.Optional[typing.Any] = None,
            params: typing.Optional[http_types.QueryParamTypes] = None,
            headers: typing.Optional[http_types.HeaderTypes] = None,
            cookies: typing.Optional[http_types.CookieTypes] = None
    ) -> typing.Any:

        with _catch_error():

            resp = await self.send_request(
                r'PATCH', url,
                content=content, data=data, files=files, json=json, params=params,
                headers=headers, cookies=cookies
            )

            return resp.body

    async def delete(
            self,
            url: http_types.URLTypes,
            *,
            content: typing.Optional[http_types.RequestContent] = None,
            data: typing.Optional[http_types.RequestData] = None,
            files: typing.Optional[http_types.RequestFiles] = None,
            json: typing.Optional[typing.Any] = None,
            params: typing.Optional[http_types.QueryParamTypes] = None,
            headers: typing.Optional[http_types.HeaderTypes] = None,
            cookies: typing.Optional[http_types.CookieTypes] = None
    ) -> typing.Any:

        with _catch_error():

            resp = await self.send_request(
                r'DELETE', url,
                content=content, data=data, files=files, json=json, params=params,
                headers=headers, cookies=cookies
            )

            return resp.body


class HTTPTextClientPool(HTTPClientPool):
    """HTTP客户端，Text模式
    """

    def _handle_response(self, response: httpx.Response) -> str:

        return response.text


class HTTPJsonClientPool(HTTPClientPool):
    """HTTP客户端，Json模式
    """

    def _handle_response(self, response: httpx.Response) -> typing.Union[typing.Dict, typing.List]:

        return response.json()


class Downloader(_ClientBase):
    """HTTP文件下载器
    """

    async def fetch(
            self,
            file: typing.BinaryIO,
            url: http_types.URLTypes,
            *,
            chunk_size: int = 65536,
            params: typing.Optional[http_types.QueryParamTypes] = None,
            headers: typing.Optional[http_types.HeaderTypes] = None,
            cookies: typing.Optional[http_types.CookieTypes] = None
    ):

        async with self._client.stream(r'GET', url, params=params, headers=headers, cookies=cookies) as response:
            async for chunk in response.aiter_bytes(chunk_size):
                file.write(chunk)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import types
import typing
import weakref
import inspect
import asyncio
import functools

import async_timeout

from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager
from contextvars import Context, ContextVar, Token

from .. import base
from ..error import catch_error
from ..cache import StackCache


def install_uvloop():
    """尝试安装uvloop
    """

    try:
        import uvloop
    except ModuleNotFoundError:
        Utils.log.warning(f'uvloop is not supported (T＿T)')
    else:
        uvloop.install()
        Utils.log.info(f'uvloop {uvloop.__version__} installed')


class Utils(base.Utils):
    """异步基础工具类

    集成常用的异步工具函数

    """

    sleep = staticmethod(asyncio.sleep)

    @staticmethod
    def is_coroutine_function(func: typing.Callable) -> bool:

        if isinstance(func, functools.partial):
            return asyncio.iscoroutinefunction(func.func)
        else:
            return asyncio.iscoroutinefunction(func)

    @staticmethod
    async def awaitable_wrapper(obj: typing.Any) -> typing.Any:
        """自适应awaitable对象
        """

        if inspect.isawaitable(obj):
            return await obj
        else:
            return obj

    @staticmethod
    @types.coroutine
    def wait_frame(count: int = 10):
        """暂停指定帧数
        """

        for _ in range(max(1, count)):
            yield

    @staticmethod
    def get_event_loop() -> asyncio.AbstractEventLoop:
        return asyncio.get_event_loop()

    @staticmethod
    def loop_time() -> float:
        """获取当前loop内部时钟
        """

        loop = asyncio.get_event_loop()

        return loop.time()

    @classmethod
    def call_soon(
            cls, callback: typing.Callable, *args,
            loop: asyncio.AbstractEventLoop = None, context: Context = None
    ) -> asyncio.Handle:
        """延时调用(能隔离上下文)
        """

        if loop is None:
            loop = asyncio.get_event_loop()

        if context is None:
            context = Context()

        return loop.call_soon(
            async_adapter(callback),
            *args,
            context=context
        )

    @classmethod
    def call_soon_threadsafe(
            cls, callback: typing.Callable, *args,
            loop: asyncio.AbstractEventLoop = None, context: Context = None
    ) -> asyncio.Handle:
        """延时调用(线程安全，能隔离上下文)
        """

        if loop is None:
            loop = asyncio.get_event_loop()

        if context is None:
            context = Context()

        return loop.call_soon_threadsafe(
            async_adapter(callback),
            *args,
            context=context
        )

    @classmethod
    def call_later(
            cls, delay: float, callback: typing.Callable, *args,
            loop: asyncio.AbstractEventLoop = None, context: Context = None
    ) -> asyncio.Handle:
        """延时指定秒数调用(能隔离上下文)
        """

        if loop is None:
            loop = asyncio.get_event_loop()

        if context is None:
            context = Context()

        return loop.call_later(
            delay,
            async_adapter(callback),
            *args,
            context=context
        )

    @classmethod
    def call_at(
            cls, when: float, callback: typing.Callable, *args,
            loop: asyncio.AbstractEventLoop = None, context: Context = None
    ) -> asyncio.Handle:
        """指定时间调用(能隔离上下文)
        """

        if loop is None:
            loop = asyncio.get_event_loop()

        if context is None:
            context = Context()

        return loop.call_at(
            when,
            async_adapter(callback),
            *args,
            context=context
        )

    @staticmethod
    def create_task(obj: typing.Union[typing.Generator, typing.Awaitable]):
        """将协程对象包装成task对象(兼容Future接口)
        """

        if inspect.isgenerator(obj) or inspect.isawaitable(obj):
            return asyncio.create_task(obj)
        else:
            return None

    @staticmethod
    def run_until_complete(future: typing.Union[typing.Coroutine, typing.Generator]):
        """运行事件循环直到future结束
        """

        loop = asyncio.get_event_loop()

        return loop.run_until_complete(future)

    @staticmethod
    @asynccontextmanager
    async def async_timeout(timeout: float):
        """异步超时等待

        async with timeout(1.5) as res:
            pass
        print(res.expired)

        """

        with catch_error():
            async with async_timeout.timeout(timeout) as res:
                yield res


def async_adapter(func: typing.Callable) -> typing.Callable:
    """异步函数适配装饰器

    使异步函数可以在同步函数中调用，即非阻塞式的启动异步函数，同时会影响上下文资源的生命周期

    """

    if not Utils.is_coroutine_function(func):
        return func

    @base.Utils.func_wraps(func)
    def _wrapper(*args, **kwargs):

        return Utils.create_task(
            func(*args, **kwargs)
        )

    return _wrapper


class WeakContextVar:
    """弱引用版的上下文资源共享器
    """

    def __init__(self, name: str):
        self._context_var: ContextVar = ContextVar(name, default=None)

    def get(self) -> typing.Any:

        ref = self._context_var.get()

        return None if ref is None else ref()

    def set(self, value: typing.Any) -> Token:
        return self._context_var.set(weakref.ref(value))


class FutureWithTimeout(asyncio.Future):
    """带超时功能的Future
    """

    def __init__(self, delay: float, default: typing.Any = None):

        super().__init__()

        self._timeout_handle = Utils.call_later(
            delay,
            self.set_result,
            default
        )

        self.add_done_callback(self._clear_timeout)

    def _clear_timeout(self, *_):

        if self._timeout_handle is not None:
            self._timeout_handle.cancel()
            self._timeout_handle = None


class FutureWithCoroutine(asyncio.Future):
    """Future实例可以被多个协程await，本类实现Future和Coroutine的桥接
    """

    def __init__(self, coroutine: typing.Coroutine):

        super().__init__()

        self._lock = asyncio.Lock()
        self._coroutine: typing.Coroutine = coroutine

    def __await__(self) -> typing.Any:
        return (yield from self._run_coroutine().__await__())

    async def _run_coroutine(self):

        async with self._lock:

            if not self.done():

                try:
                    self.set_result(
                        await self._coroutine
                    )
                except Exception as err:
                    super().cancel()
                    raise err

        return Utils.deepcopy(self.result())

    def cancel(self, *args, **kwargs):

        super().cancel()

        self._coroutine.close()


class AsyncCirculatory:
    """异步循环器

    提供一个循环体内的代码重复执行管理逻辑，可控制超时时间、执行间隔(LoopFrame)和最大执行次数

    async for index in AsyncCirculatory():
        pass

    其中index为执行次数，从1开始

    """

    def __init__(
            self, timeout: typing.Union[int, float] = 0,
            interval: typing.Union[int, float] = 0xff,
            max_times: int = 0
    ):

        self._expire_time: typing.Union[int, float] = 0

        if timeout > 0:
            self._expire_time = Utils.loop_time() + timeout

        self._interval: typing.Union[int, float] = interval
        self._max_times: int = max_times

        self._current: int = 0

    def __aiter__(self):
        return self

    async def __anext__(self):

        if self._current > 0:

            if (self._max_times > 0) and (self._max_times <= self._current):
                raise StopAsyncIteration()

            if (self._expire_time > 0) and (self._expire_time <= Utils.loop_time()):
                raise StopAsyncIteration()

            await self._sleep()

        self._current += 1

        return self._current

    async def _sleep(self):
        await Utils.wait_frame(self._interval)


class AsyncCirculatoryForSecond(AsyncCirculatory):

    def __init__(
            self, timeout: typing.Union[int, float] = 0,
            interval: typing.Union[int, float] = 0xff,
            max_times: int = 0
    ):
        super().__init__(timeout, interval, max_times)

    async def _sleep(self):
        await Utils.sleep(self._interval)


class FuncWrapper(base.FuncWrapper):
    """非阻塞异步函数包装器

    将多个同步或异步函数包装成一个可调用对象

    """

    def __call__(self, *args, **kwargs):

        for func in self._callables:
            Utils.call_soon(func, *args, **kwargs)


class AsyncConstructor(metaclass=ABCMeta):
    """实现了__async_init__异步构造函数
    """

    def __await__(self):

        yield from self.__async_init__().__await__()

        return self

    @abstractmethod
    async def __async_init__(self):
        raise NotImplementedError()


class AsyncFuncWrapper(base.FuncWrapper):
    """阻塞式异步函数包装器

    将多个同步或异步函数包装成一个可调用对象

    """

    async def __call__(self, *args, **kwargs):

        for func in self._callables:

            with catch_error():
                await Utils.awaitable_wrapper(
                    func(*args, **kwargs)
                )


class ShareFuture:
    """共享Future装饰器

    同一时刻并发调用函数时，使用该装饰器的函数签名一致地调用，会共享计算结果

    """

    def __init__(self):
        self._future: typing.Dict[str, FutureWithCoroutine] = {}

    def __call__(self, func: typing.Callable) -> typing.Callable:

        @base.Utils.func_wraps(func)
        async def _wrapper(*args, **kwargs):

            nonlocal self, func

            return await self._make_future(
                func, *args, **kwargs
            )

        return _wrapper

    def _make_future(self, func: typing.Callable, *args, **kwargs) -> FutureWithCoroutine:

        func_sign = Utils.params_sign(func, *args, **kwargs)
        future = self._future.get(func_sign)

        if not future:

            future = self._future[func_sign] = FutureWithCoroutine(func(*args, **kwargs))

            future.add_done_callback(
                Utils.func_partial(self._clear_future, func_sign)
            )

        return future

    def _clear_future(self, func_sign: str, _: typing.Any):

        if func_sign in self._future:
            del self._future[func_sign]


class FuncCache(ShareFuture):
    """函数缓存

    使用堆栈缓存实现的函数缓存，在有效期内函数签名一致就会命中缓存

    """

    def __init__(self, maxsize: int = 0xffff, ttl: float = 10):

        super().__init__()

        self._cache = StackCache(maxsize, ttl)

    def __call__(self, func: typing.Callable) -> typing.Callable:

        @base.Utils.func_wraps(func)
        async def _wrapper(*args, **kwargs):

            nonlocal self, func

            return await self._make_future(
                self._do_func_with_cache, func, *args, **kwargs
            )

        return _wrapper

    async def _do_func_with_cache(self, func: typing.Callable, *args, **kwargs) -> typing.Any:

        func_sign = Utils.params_sign(func, *args, **kwargs)

        result = self._cache.get(func_sign)

        if result is None:

            result = await func(*args, **kwargs)

            if result is not None:
                self._cache.set(func_sign, result)

        return result



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import typing
import aiormq
import aio_pika

from abc import abstractmethod

from ...extend.asyncio.pool import ObjectInterface, ObjectPool


class _ProducerBase(ObjectInterface):

    def __init__(self, connection: aio_pika.abc.AbstractRobustConnection):

        self._connection: aio_pika.abc.AbstractRobustConnection = connection
        self._channel: typing.Optional[aio_pika.abc.AbstractRobustChannel] = None
        self._exchange: typing.Optional[aio_pika.abc.AbstractExchange] = None

    @abstractmethod
    async def open(self, *args, **kwargs):

        raise NotImplementedError()

    async def close(self, with_connection: bool = False):

        await self._channel.close()

        if with_connection is True:
            await self._connection.close()

    async def publish(
            self,
            message: typing.Union[bytes, aio_pika.Message],
            routing_key: str = r'',
            **kwargs
    ) -> aiormq.abc.ConfirmationFrameType:

        return await self._exchange.publish(
            message if isinstance(message, aio_pika.Message) else aio_pika.Message(message),
            routing_key, **kwargs
        )


class RabbitMQProducer(_ProducerBase):
    """RabbitMQ发布者
    """

    async def open(
            self, *,
            channel_number: typing.Optional[int] = None,
            publisher_confirms: bool = True,
            on_return_raises: bool = False
    ):

        self._channel = await self._connection.channel(channel_number, publisher_confirms, on_return_raises)
        self._exchange = self._channel.default_exchange


class RabbitMQProducerForExchange(_ProducerBase):
    """RabbitMQ交换机发布者
    """

    def __init__(self, connection: aio_pika.abc.AbstractRobustConnection, exchange_name: str):

        super().__init__(connection)

        self._exchange_name: str = exchange_name

    async def open(
            self, *,
            exchange_type: aio_pika.ExchangeType = aio_pika.ExchangeType.FANOUT,
            exchange_config: typing.Optional[typing.Dict] = None,
            channel_number: int = None,
            publisher_confirms: bool = True,
            on_return_raises: bool = False
    ):

        if exchange_config is None:
            exchange_config = {}

        self._channel = await self._connection.channel(channel_number, publisher_confirms, on_return_raises)
        self._exchange = await self._channel.declare_exchange(self._exchange_name, exchange_type, **exchange_config)

##################################################

class _ProducerPoolBase(ObjectPool):

    def __init__(self, pool_size: int, connection: aio_pika.abc.AbstractRobustConnection):

        super().__init__(pool_size)

        self._connection = connection

    @abstractmethod
    def _create(self) -> ObjectInterface:

        raise NotImplementedError()

    async def close(self, with_connection: bool = False):

        await super().close()

        if with_connection is True:
            await self._connection.close()

    async def publish(
            self,
            message: typing.Union[bytes, aio_pika.Message],
            routing_key: str = r'',
            **kwargs
    ) -> aiormq.abc.ConfirmationFrameType:

        async with self.get() as producer:
            return await producer.publish(
                message if isinstance(message, aio_pika.Message) else aio_pika.Message(message),
                routing_key, **kwargs
            )


class RabbitMQProducerPool(_ProducerPoolBase):
    """RabbitMQ发布者连接池
    """

    def _create(self) -> RabbitMQProducer:

        return RabbitMQProducer(self._connection)

    async def open(
            self, *,
            channel_number: typing.Optional[int] = None,
            publisher_confirms: bool = True,
            on_return_raises: bool = False
    ):

        await self._connection.ready()

        await super().open(
            channel_number=channel_number,
            publisher_confirms=publisher_confirms,
            on_return_raises=on_return_raises
        )


class RabbitMQProducerForExchangePool(_ProducerPoolBase):
    """RabbitMQ交换机发布者连接池
    """

    def __init__(self, pool_size: int, connection: aio_pika.abc.AbstractRobustConnection, exchange_name: str):

        super().__init__(pool_size, connection)

        self._exchange_name: str = exchange_name

    def _create(self) -> RabbitMQProducerForExchange:

        return RabbitMQProducerForExchange(self._connection, self._exchange_name)

    async def open(
            self, *,
            exchange_type: aio_pika.ExchangeType = aio_pika.ExchangeType.FANOUT,
            exchange_config: typing.Optional[typing.Dict] = None,
            channel_number: typing.Optional[int] = None,
            publisher_confirms: bool = True,
            on_return_raises: bool = False
    ):

        await self._connection.ready()

        await super().open(
            exchange_type=exchange_type,
            exchange_config=exchange_config,
            channel_number=channel_number,
            publisher_confirms=publisher_confirms,
            on_return_raises=on_return_raises
        )



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from abc import abstractmethod, ABC
from typing import Any, Union, Literal, Optional, Iterable, Tuple

from coredis import typing, Redis, RedisCluster
from coredis.recipes.locks import LuaLock as _LuaLock
from coredis.pool import ConnectionPool as _ConnectionPool, ClusterConnectionPool as _ClusterConnectionPool

from .base import Utils

from ..error import catch_error, catch_warning
from ..interface import AsyncContextManager
from .transaction import Transaction


class ConnectionPool(_ConnectionPool):

    @property
    def current_connections(self) -> int:
        return self._created_connections


class ClusterConnectionPool(_ClusterConnectionPool):

    @property
    def current_connections(self) -> int:
        return sum(self._created_connections_per_node.values())


class LuaLock(AsyncContextManager, _LuaLock):

    def __init__(
            self, client: Any, name: str,
            timeout: Optional[float] = None, sleep: float = 0.1,
            blocking: bool = True, blocking_timeout: Optional[float] = None
    ):
        _LuaLock.__init__(self, client, name, timeout, sleep, blocking, blocking_timeout)

    async def _context_release(self):
        with catch_warning():
            await self.release()

    async def acquire(self) -> bool:
        if self.local.get() is None:
            return await super().acquire()
        else:
            return await super().extend(self.timeout)

    async def release(self):
        if self.local.get() is not None:
            await super().release()


class _PoolMixin(ABC):

    connection_pool: Union[ConnectionPool, ClusterConnectionPool]

    def __init__(self, min_connections=0):

        self._name: str = Utils.uuid1()[:8]
        self._key_prefix: Optional[str] = None
        self._min_connections: int = min_connections if min_connections else 0

    @abstractmethod
    async def _init_connection(self):
        raise NotImplementedError()

    @abstractmethod
    async def get(self, *args, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    async def getset(self, *args, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    async def set(self, *args, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    async def setex(self, *args, **kwargs):
        raise NotImplementedError()

    def set_key_prefix(self, value: str):

        self._key_prefix = value

    def get_safe_key(self, key, *args, **kwargs) -> str:

        if self._key_prefix:
            _key = f'{self._key_prefix}:{key}'
        else:
            _key = key

        if args or kwargs:
            _key = f'{_key}:{Utils.params_sign(*args, **kwargs)}'

        return _key

    async def close(self):

        self.connection_pool.disconnect()

    def allocate_lock(
            self, name: str,
            timeout: Optional[float] = None, sleep: float = 0.1,
            blocking: bool = True, blocking_timeout: Optional[float] = None
    ) -> LuaLock:

        return LuaLock(self, name, timeout, sleep, blocking, blocking_timeout)

    async def get_obj(self, name: str) -> Any:

        result = await self.get(name)

        return Utils.pickle_loads(result) if result else result

    async def getset_obj(self, name: str, value: Any) -> Any:

        _value = Utils.pickle_dumps(value)

        result = await self.getset(name, _value)

        return Utils.pickle_loads(result) if result else result

    async def set_obj(self, name: str, value: Any, seconds: Optional[float] = None):

        _value = Utils.pickle_dumps(value)

        if seconds is None:
            return await self.set(name, _value)
        else:
            return await self.setex(name, _value, seconds)


class RedisPool(Redis, _PoolMixin):
    """StrictRedis连接管理
    """

    connection_pool: ConnectionPool

    def __init__(
            self,
            host:str, port:int, db: int = 0,
            username: Optional[str] = None, password: Optional[str] = None,
            min_connections: int = 0, max_connections: int =32,
            max_idle_time: float = 43200, idle_check_interval: float = 1,
            protocol_version: Literal[2, 3] = 2,
            **kwargs
    ):

        Redis.__init__(
            self,
            host, port, db,
            username=username, password=password, max_connections=max_connections,
            max_idle_time=max_idle_time, idle_check_interval=idle_check_interval,
            protocol_version=protocol_version, connection_pool_cls=ConnectionPool,
            **kwargs
        )

        _PoolMixin.__init__(self, min(min_connections, max_connections))

    async def _init_connection(self):

        connections = []

        with catch_error():
            for _ in range(self._min_connections):
                connection = await self.connection_pool.get_connection()
                connections.append(connection)

        with catch_error():
            for connection in connections:
                self.connection_pool.release(connection)

    async def initialize(self):

        await self._init_connection()

        config = self.connection_pool.connection_kwargs

        Utils.log.info(
            f"Redis Pool ({self._name}) [{config[r'host']}:{config[r'port']}] initialized: "
            f"{self.connection_pool.current_connections}/{self.connection_pool.max_connections}"
        )

        return self


class RedisClusterPool(RedisCluster, _PoolMixin):
    """StrictRedisCluster连接管理
    """

    connection_pool: ClusterConnectionPool

    def __init__(
            self,
            host: str, port: int, startup_nodes: Optional[Iterable[typing.Node]] = None,
            username: Optional[str] = None, password: Optional[str] = None,
            min_connections: int = 0, max_connections: int = 32,
            max_idle_time: float = 43200, idle_check_interval: float = 1,
            protocol_version: Literal[2, 3] = 2,
            **kwargs
    ):

        RedisCluster.__init__(
            self,
            host, port, startup_nodes=startup_nodes,
            username=username, password=password, max_connections=max_connections,
            max_idle_time=max_idle_time, idle_check_interval=idle_check_interval,
            protocol_version=protocol_version, connection_pool_cls=ClusterConnectionPool,
            **kwargs
        )

        _PoolMixin.__init__(self, min(min_connections, max_connections))

    async def _init_connection(self):

        connections = []

        with catch_error():
            for _ in range(self._min_connections):
                connection = await self.connection_pool.get_connection_by_key(Utils.uuid1())
                connections.append(connection)

        with catch_error():
            for connection in connections:
                self.connection_pool.release(connection)

    async def initialize(self):

        await self.connection_pool.initialize()

        await self._init_connection()

        nodes = self.connection_pool.nodes.nodes

        Utils.log.info(
            f"Redis Cluster Pool ({self._name}) {list(nodes.keys())} initialized: "
            f"{self.connection_pool.current_connections}/{self.connection_pool.max_connections}"
        )

        return self


class RedisDelegate:
    """Redis功能组件
    """

    def __init__(self):

        self._redis_pool: Optional[RedisPool, RedisClusterPool] = None

    @property
    def redis_pool(self) -> Union[RedisPool, RedisClusterPool]:

        return self._redis_pool

    async def init_redis_single(
            self,
            host:str, port:int, db: int = 0,
            username: Optional[str] = None, password: Optional[str] = None, *,
            min_connections: int = 0, max_connections: int =32,
            max_idle_time: float = 43200, idle_check_interval: float = 1,
            protocol_version: Literal[2, 3] = 2,
            **kwargs
    ):

        self._redis_pool = await RedisPool(
            host, port, db, username, password,
            min_connections, max_connections,
            max_idle_time, idle_check_interval,
            protocol_version,
            **kwargs
        ).initialize()

        return self._redis_pool

    async def init_redis_cluster(
            self,
            host: str, port: int, startup_nodes: Optional[Iterable[typing.Node]] = None,
            username: Optional[str] = None, password: Optional[str] = None, *,
            min_connections: int = 0, max_connections: int = 32,
            max_idle_time: float = 43200, idle_check_interval: float = 1,
            protocol_version: Literal[2, 3] = 2,
            **kwargs
    ):

        self._redis_pool = await RedisClusterPool(
            host, port, startup_nodes, username, password,
            min_connections, max_connections,
            max_idle_time, idle_check_interval,
            protocol_version,
            **kwargs
        ).initialize()

        return self._redis_pool

    def set_redis_key_prefix(self, value: str):

        self._redis_pool.set_key_prefix(value)

    async def close_redis_pool(self):

        await self._redis_pool.close()

    def get_redis_client(self) -> Union[RedisPool, RedisClusterPool]:

        return self._redis_pool


class ShareCache(AsyncContextManager):
    """共享缓存，使用with进行上下文管理

    基于分布式锁实现的一个缓存共享逻辑，保证在分布式环境下，同一时刻业务逻辑只执行一次，其运行结果会通过缓存被共享

    """

    def __init__(
            self, redis_client: Union[RedisPool, RedisClusterPool], share_key: str,
            lock_timeout: float = 60, lock_blocking_timeout: float = 60
    ):

        self._redis_client: Union[RedisPool, RedisClusterPool] = redis_client
        self._share_key: str = redis_client.get_safe_key(share_key)

        self._lock: LuaLock = self._redis_client.allocate_lock(
            redis_client.get_safe_key(f'share_cache:{share_key}'),
            timeout=lock_timeout, blocking=True, blocking_timeout=lock_blocking_timeout
        )

        self.result = None

    async def _context_release(self):

        await self.release()

    async def get(self) -> Any:

        result = await self._redis_client.get_obj(self._share_key)

        if result is None:

            if await self._lock.acquire():
                result = await self._redis_client.get_obj(self._share_key)

        return result

    async def set(self, value: Any, expire: Optional[float] = None):

        return await self._redis_client.set_obj(self._share_key, value, expire)

    async def release(self):

        if self._lock:
            await self._lock.release()

        self._redis_client = self._lock = None


class PeriodCounter:

    MIN_EXPIRE = 60

    def __init__(self, redis_client: Union[RedisPool, RedisClusterPool], key_prefix: str, time_slice: int):

        self._redis_client: Union[RedisPool, RedisClusterPool] = redis_client

        self._key_prefix = key_prefix
        self._time_slice = time_slice

    def _get_key(self) -> str:

        time_period = Utils.math.floor(Utils.timestamp() / self._time_slice)

        return self._redis_client.get_safe_key(f'{self._key_prefix}:{time_period}')

    async def _execute(self, key: str, val: int) -> int:

        async with await self._redis_client.pipeline(transaction=False) as pipeline:

            await pipeline.incrby(key, val)
            await pipeline.expire(key, max(self._time_slice, self.MIN_EXPIRE))

            res, _ = await pipeline.execute()

            return res

    async def incr(self, val: int = 1):

        return await self._execute(self._get_key(), val)

    async def incr_with_trx(self, val: int = 1) -> Tuple[int, Transaction]:

        _key = self._get_key()

        res = await self._execute(_key, val)

        if res is not None:
            trx = Transaction()
            trx.add_rollback_callback(self._execute, _key, -val)
        else:
            trx = None

        return res, trx

    async def decr(self, val: int = 1) -> int:

        return await self._execute(self._get_key(), -val)

    async def decr_with_trx(self, val: int = 1) -> Tuple[int, Transaction]:

        _key = self._get_key()

        res = await self._execute(_key, -val)

        if res is not None:
            trx = Transaction()
            trx.add_rollback_callback(self._execute, _key, val)
        else:
            trx = None

        return res, trx
